use super::Solution;
use crate::utils::Direction;
use std::collections::{BinaryHeap, HashMap};

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
struct State {
    cost: u32,
    pos: (usize, usize),
    dir: Direction,
    steps: u8,
}

impl Ord for State {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        other
            .cost
            .cmp(&self.cost)
            .then_with(|| self.pos.cmp(&other.pos))
    }
}

impl PartialOrd for State {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
struct StateKey {
    pos: (usize, usize),
    dir: Direction,
    steps: u8,
}

impl From<State> for StateKey {
    fn from(value: State) -> Self {
        Self {
            pos: value.pos,
            dir: value.dir,
            steps: value.steps,
        }
    }
}

pub struct Day17 {}

impl Solution for Day17 {
    fn part1(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let mut grid: HashMap<(usize, usize), u32> = HashMap::new();

        for (y, row) in input.iter().enumerate() {
            for (x, col) in row.chars().enumerate() {
                grid.insert((x, y), col.to_digit(10).unwrap());
            }
        }

        Ok(Box::new(
            self.dijkstra(
                (0, 0),
                (input[0].len() - 1, input.len() - 1),
                (input[0].len(), input.len()),
                1,
                3,
                grid,
            )
            .unwrap(),
        ))
    }

    fn part2(
        &self,
        input: &mut Vec<String>,
    ) -> Result<Box<dyn std::fmt::Display + Sync>, Box<dyn std::error::Error>> {
        let mut grid: HashMap<(usize, usize), u32> = HashMap::new();

        for (y, row) in input.iter().enumerate() {
            for (x, col) in row.chars().enumerate() {
                grid.insert((x, y), col.to_digit(10).unwrap());
            }
        }

        Ok(Box::new(
            self.dijkstra(
                (0, 0),
                (input[0].len() - 1, input.len() - 1),
                (input[0].len(), input.len()),
                4,
                10,
                grid,
            )
            .unwrap(),
        ))
    }

    fn get_day(&self) -> u8 {
        17
    }
}

impl Day17 {
    fn dijkstra(
        &self,
        start: (usize, usize),
        end: (usize, usize),
        lens: (usize, usize),
        min_step: u8,
        max_step: u8,
        grid: HashMap<(usize, usize), u32>,
    ) -> Option<u32> {
        let mut dist = HashMap::<StateKey, u32>::new();
        let mut heap = BinaryHeap::new();
        let st1 = State {
            cost: 0,
            pos: start,
            dir: Direction::Right,
            steps: 0,
        };
        let st2 = State {
            cost: 0,
            pos: start,
            dir: Direction::Down,
            steps: 0,
        };
        dist.insert(st1.into(), 0);
        dist.insert(st2.into(), 0);
        heap.push(st1);
        heap.push(st2);

        while let Some(
            state @ State {
                cost,
                pos,
                dir,
                steps,
            },
        ) = heap.pop()
        {
            if pos == end && steps >= min_step {
                return Some(cost);
            }
            if dist.get(&state.into()).is_some_and(|&c| c < cost) {
                continue;
            }
            for (n, d) in self.neighbours(pos, dir, lens) {
                let next = State {
                    cost: cost + grid[&n],
                    pos: n,
                    dir: d,
                    steps: if d == dir { steps + 1 } else { 1 },
                };
                if next.steps > max_step || dist.get(&next.into()).is_some_and(|&c| c <= next.cost)
                {
                    // For p1 and p2
                    // Streak gets too long
                    // Or a better solution/path already exists
                    continue;
                }
                if next.dir != dir && steps < min_step {
                    // For p2
                    // Making a turn but previous streak is too short
                    continue;
                }
                heap.push(next);
                dist.insert(next.into(), next.cost);
            }
        }
        None
    }

    fn neighbours(
        &self,
        (x, y): (usize, usize),
        dir: Direction,
        (x_len, y_len): (usize, usize),
    ) -> Vec<((usize, usize), Direction)> {
        const fn deltas(x: usize, y: usize) -> [((usize, usize), Direction); 4] {
            [
                ((x, y.saturating_sub(1)), Direction::Up),
                ((x, y + 1), Direction::Down),
                ((x.saturating_sub(1), y), Direction::Left),
                ((x + 1, y), Direction::Right),
            ]
        }

        deltas(x, y)
            .into_iter()
            .filter_map(|(n, d)| {
                if n == (x, y) || n.0 > x_len - 1 || n.1 > y_len - 1 || d == dir.flip() {
                    None
                } else {
                    Some((n, d))
                }
            })
            .collect()
    }
}

/// Test from puzzle input
#[cfg(test)]
mod test {
    use super::*;
    use crate::*;

    #[test]
    fn part1() {
        let challenge = day17::Day17 {};

        //Complete the Challenge
        let answer = challenge
            .part1(
                utils::get_input(challenge.get_day(), utils::InputType::Test1)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "102");
    }

    #[test]
    fn part2() {
        let challenge = day17::Day17 {};

        //Complete the Challenge
        let answer = challenge
            .part2(
                utils::get_input(challenge.get_day(), utils::InputType::Test2)
                    .unwrap()
                    .as_mut(),
            )
            .unwrap()
            .to_string();

        assert_eq!(answer, "94");
    }
}
